import cls from 'classnames';
import debounce from 'lodash.debounce';
import classes from './MarginalAnnotations.module.scss';
import { AnnotationBody } from './AnnotationBody';
import { Component, h } from 'preact';
import { getSingleCreatorName } from 'web-annotation-utils';
import type { WebAnnotation } from 'web-annotation-utils';
import type { IAnnotation, IAnnotationWithSource } from '../storage/Annotation';

export interface MarginalAnnotationCardProps {
  annotation: IAnnotationWithSource;
  anchorPosition?: number | undefined;
  extraClasses?: string;
  inFocus?: boolean;
  onUpdateAnnotation: (
    id: IAnnotation['_id'],
    newWebAnnotation: WebAnnotation,
  ) => Promise<void>;
  onDeleteAnnotation: (id: IAnnotation['_id']) => Promise<void>;
  [other: string]: any;
}

interface MarginalAnnotationCardState {}

export class MarginalAnnotationCard extends Component<
  MarginalAnnotationCardProps,
  MarginalAnnotationCardState
> {
  state = {};

  onBodyChange = debounce(
    async (newBody: WebAnnotation['body']) => {
      const newWebAnnotation: WebAnnotation = {
        ...this.props.annotation.annotation,
        bodyValue: undefined,
        body: newBody,
        modified: new Date().toISOString(),
      };
      await this.props.onUpdateAnnotation(
        this.props.annotation._id,
        newWebAnnotation,
      );
    },
    1_000,
    { maxWait: 10_000 },
  );

  render(
    {
      annotation,
      anchorPosition,
      extraClasses,
      inFocus,
      ...otherProps
    }: MarginalAnnotationCardProps,
    {}: MarginalAnnotationCardState,
  ) {
    const editable = !!annotation.source.writable;
    const anchorPositionStyle =
      anchorPosition !== undefined
        ? {
            style: { marginTop: anchorPosition },
          }
        : {};
    const annotationCard = (
      <div
        class={classes.annotationWrapper}
        {...anchorPositionStyle}
        key={annotation._id}
      >
        <aside
          class={cls(
            classes.annotation,
            { [classes.dirty]: annotation.dirty, [classes.editable]: editable },
            extraClasses,
          )}
          {...otherProps}
        >
          <AnnotationBody
            body={annotation.annotation.body}
            bodyValue={annotation.annotation.bodyValue}
            editable={editable}
            onChange={(newBody) => this.onBodyChange(newBody)}
            onBlur={() => this.onBodyChange.flush()}
            inFocus={inFocus}
          />
          <div class={classes.cardBottom}>
            <span class={classes.creator}>
              {getSingleCreatorName(annotation.annotation)}
            </span>
            <span>
              {annotation.dirty && (
                <span
                  title={`Your edits on this annotation have not yet been stored; if this problem persists, try manually refresh its annotation collection (“${annotation.source?.title}”).`}
                >
                  🖉{' '}
                </span>
              )}
              {editable && (
                <button
                  onClick={() =>
                    this.props.onDeleteAnnotation(this.props.annotation._id)
                  }
                  class={classes.button}
                  title="Delete this annotation"
                >
                  🗑
                </button>
              )}
            </span>
          </div>
        </aside>
      </div>
    );

    // If it has a specified vertical position, wrap it in a separate absolutely positioned container
    if (anchorPosition !== undefined) {
      return (
        <div
          class={classes.singleAnnotationOuterContainer}
          key={annotation._id}
        >
          <div class={cls(classes.singleAnnotationContainer)}>
            {annotationCard}
          </div>
        </div>
      );
    } else {
      return annotationCard;
    }
  }
}
